/*
 * $Log: cleanup.c,v $
 * Revision 1.2  2004/07/14 05:09:47  mi
 * Prepare for the 1.0 release
 *
 * Revision 1.1  2004/06/29 20:55:33  mi
 * Initial revision
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <libmilter/mfapi.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "skem.h"

const char *cleanup_version = "$Id: cleanup.c,v 1.2 2004/07/14 05:09:47 mi Exp mi $"
	" compiled on " __DATE__ " at "__TIME__;

void *
skem_cleanup(void *direct)
{
	DIR		*dir;
	struct dirent	*entry;
	int	 	 result;
	unsigned	 select_failures;
	struct stat	 sb;
	struct timeval	 current;
	char		 mtime[26];

	if (O.size != sizeof(O)) {
		syslog(LOG_ERR, "Options structure sizes do not match.");
		if (direct)
			return (void *)EX_SOFTWARE;
		else {
			mi_stop_milters(MILTER_ABRT);
			return NULL;
		}
	}
	syslog(LOG_DEBUG, "periodic cleanup: period %ld seconds%s, "
	    "permanent bans removed after %u seconds since last seen",
	    O.cleanupperiod.tv_sec,
	    O.cleanupperiod.tv_sec ? "" : " (one shot)",
	    O.permcleanup);
	for (result = 0, select_failures = 0;
	    /* forever */;
	    /*
	     * We drop the constness from O because we know, select(2) will
	     * not modify it, even though the compiler does not :-(
	     */
	    result = select(0, 0, 0, 0, (struct timeval *)&O.cleanupperiod)) {
		if (result != 0) {
			select_failures++;
			syslog(LOG_WARNING, "Cleanup thread: %m (failure %u)",
			    select_failures);
			if (select_failures > 50) {
				syslog(LOG_ERR, "Too many select failures in "
				    "a row. Terminating cleanup thread");
				return (void *)EX_OSERR;
			}
			continue;
		}
		if (select_failures) {
			syslog(LOG_DEBUG, "Resetting the subsequent failures "
			    "counter from %u back to 0", select_failures);
			select_failures = 0;
		}

		/* syslog(LOG_DEBUG, "Cleanup time"); */
		dir = opendir(".");
		if (dir == NULL) {
			syslog(LOG_ERR, "Can't open the current directory "
			    "for cleanup: %m");
			continue;
		}
		if (gettimeofday(&current, NULL)) {
			syslog(LOG_ERR, "Can't obtain the current time "
			    "for cleanup: %m");
			continue;
		}
		while ((entry = readdir(dir))) {
			if ((entry->d_type != DT_LNK &&
			    entry->d_type != DT_REG)
			    || !skem_isanIP(entry->d_name)
			    || lstat(entry->d_name, &sb) /* skip quietly */
			    || sb.st_mode & S_ISUID /* white list */
			    || (sb.st_mode & S_ISGID && /* black list */
				current.tv_sec - sb.st_ctime < O.permcleanup)
			    || current.tv_sec - sb.st_mtime < O.grace) {
#if 0
				syslog(LOG_DEBUG, "skipping %s", entry->d_name);
#endif
				continue;
			}

			ctime_r(&sb.st_mtime, mtime);
			if (sb.st_mode & S_ISGID) {
				char	ctime[26];

				ctime_r(&sb.st_ctime, ctime);
				syslog(LOG_DEBUG, "The permanent ban on %s"
				    " was created on %.24s, but not used"
				    " since %.24s. Removing", entry->d_name,
				    mtime, ctime);
			} else
				syslog(LOG_DEBUG, "Removing the expired ban "
				    "on %s (from %.24s)",
				    entry->d_name, mtime);

			if (unlink(entry->d_name))
				syslog(LOG_ERR, "Can't unlink %s in "
				    "cleanup: %m", entry->d_name);

		}

		if (closedir(dir))
			syslog(LOG_WARNING, "closedir: %m");

		if (O.cleanupperiod.tv_sec == 0)
			break;
	}
	syslog(LOG_DEBUG, "cleanup: exiting");
	return NULL;
}
