/*
 * $Log: logwatcher.c,v $
 * Revision 1.2  2004/07/14 05:09:47  mi
 * Prepare for the 1.0 release
 *
 * Revision 1.1  2004/06/29 20:55:33  mi
 * Initial revision
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <libmilter/mfapi.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <regex.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "skem.h"

const char *logwatch_version = "$Id: logwatcher.c,v 1.2 2004/07/14 05:09:47 mi Exp mi $"
    " compiled on " __DATE__ " at "__TIME__;

int	mustexit = 0;

void *
skem_logwatch(void *input /* unused */)
{
	regex_t lp;
	int	errs, result, len;
	char	buf[1024], *p, e;
	const char *rejection;
	regmatch_t ipboundaries[2];
	unsigned long long matches = 0;
	struct	stat sb;

	if (O.size != sizeof(O)) {
		syslog(LOG_ERR, "Options structure sizes do not match.");
		mi_stop_milters(MILTER_ABRT);
		return NULL;
	}

	errs = regcomp(&lp, O.ippattern, REG_EXTENDED);
	if (errs) {

		regerror(errs, &lp, buf, sizeof buf);
		syslog(LOG_ERR, "Can't compile regular expression %s: %s",
		    O.ippattern, buf);
		mi_stop_milters(MILTER_ABRT);
		return NULL;
	}

	if (lp.re_nsub != 1) {
		syslog(LOG_ERR, "There should be exactly ONE parenthesized "
		    "expression in the pattern -- the IP address. Your ``%s'' has %ld",
		    O.ippattern, (long)lp.re_nsub);
		mi_stop_milters(MILTER_ABRT);
		return NULL;
	}

	if (input == NULL)
		input = stdin;

	for (;;) {
		int	mode = 0;

		if (fgets(buf, sizeof buf, (FILE *)input) == NULL) {
			if (feof((FILE *)input)) {
				syslog(LOG_INFO, "logwatch: EOF on input. "
				    "Exiting after %lld matches", matches);
				mi_stop_milters(MILTER_STOP);
				return NULL;
			}
			errs++;
			syslog(LOG_WARNING, "logwatch: error #%d: %m", errs);
			if (errs < 50)
				continue;
			syslog(LOG_ERR, "logwatch: too many errors. Exiting");
			mi_stop_milters(MILTER_ABRT);
			return NULL;
		} else if (errs) {
			syslog(LOG_WARNING, "logwatch: resetting error counter "
			    "from %d back to 0", errs);
			errs = 0;
		}

		/*
		 * Nil-terminate the buffer, because fgets() may not do it
		 * if the buffer is not big enough.
		 * Would use FreeBSD's fgetln() or Linux's getline() here,
		 * but neither is portable...
		 */
		buf[sizeof buf - 1] = '\0';

		/* Quickly check for the rejection string: */
		if (O.reject) {
			if ((rejection = strstr(buf, O.reject)) == NULL)
				continue;
		} else
			rejection = buf;

		/* See, if the string contains the spamword, if given */
		if (O.spamword && strstr(buf, O.spamword))
			mode = S_ISGID;

		result = regexec(&lp, buf, 2, ipboundaries, 0);
		switch (result) {
		case 0:
			matches++;
			break;
		case REG_NOMATCH:
			continue;
		case REG_ESPACE:
			/* Better be short here, out of memory already */
			syslog(LOG_CRIT, "Out of memory");
			mi_stop_milters(MILTER_ABRT);
			return NULL;
		default:
			syslog(LOG_ERR, "logwatch: troublesome string ``%s''", buf);
			regerror(result, &lp, buf, sizeof buf);
			syslog(LOG_ERR, "Internal error in regex-functionality: "
			    "%d: %s", result, buf);
			if (result == REG_ASSERT) {
				mi_stop_milters(MILTER_ABRT);
				return NULL;
			}
		}
		/*
		 * The 0th match is the whole string, so we only look the 1st
		 */
		if (ipboundaries[1].rm_eo == -1)
			continue;
		len = ipboundaries[1].rm_eo - ipboundaries[1].rm_so;
		if (len == 0) {
			syslog(LOG_WARNING, "An empty string matched in "
			    "``%s'' at %d", buf,
			    (int)ipboundaries[result].rm_so);
			continue;
		}
		p = buf + ipboundaries[1].rm_so;
		e = p[len];
		p[len] = '\0';

		if (!skem_isanIP(p)) {
			p[len] = e;
			syslog(LOG_WARNING, "logwatch: the ``%.*s'' "
			    "substring of ``%s'' matched, "
			    "but it is not an IP-address", len, p, buf);
			continue;
		}

		if (lstat(p, &sb) == 0) {
			syslog(LOG_DEBUG, "logwatcher: %s already listed", p);
			continue;
		}
		if (skem_touch(p, mode) == -1)
			syslog(LOG_WARNING, "logwatcher: failed to list %s:"
			    " %m", p);
		else {
			char *pb;

			p[len] = e;
			pb = strchr(rejection, '\n');
			if (pb)
				*pb = '\0';
			syslog(LOG_INFO, "logwatcher:%s banned %.*s: %s",
			    mode == S_ISGID ? " permanently" : "", len, p,
			    rejection);
		}

	}

	return NULL;
}
